﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Windows.Forms;
using System.Reflection;

namespace CustomPropertyGrid
{
	public class MemberwisePropertyEditor : GroupedPropertyEditor
	{
		[Flags]
		public enum MemberFlags
		{
			None		= 0x0,

			Properties	= 0x1,
			Fields		= 0x2,

			All			= Properties | Fields,
			Default		= All
		}

		private	MemberFlags	flags	= MemberFlags.Default;

		public MemberwisePropertyEditor(PropertyEditor parentEditor, PropertyGrid parentGrid, MemberFlags flags) : base(parentEditor, parentGrid)
		{
			this.flags = flags;
			this.Header.ResetClicked += new EventHandler(Header_ResetClicked);
		}

		public override void InitContent()
		{
			base.InitContent();

			this.ClearPropertyEditors();
			if (this.EditedType != null)
			{
				// Generate and add property editors for the current type
				this.BeginUpdate();
				if ((this.flags & MemberFlags.Properties) != MemberFlags.None)
				{
					PropertyInfo[] propArr = this.EditedType.GetProperties(BindingFlags.Instance | BindingFlags.Public);
					foreach (PropertyInfo prop in propArr)
					{
						if (!prop.CanRead) continue;
						if (prop.GetIndexParameters().Length > 0) continue;
						PropertyEditor e = this.ParentGrid.PropertyEditorProvider.CreateEditor(prop.PropertyType, this, this.ParentGrid);
						e.Getter = this.CreatePropertyValueGetter(prop);
						e.Setter = prop.CanWrite ? this.CreatePropertyValueSetter(prop) : null;
						e.PropertyName = prop.Name;
						this.AddPropertyEditor(e);
					}
				}
				if ((this.flags & MemberFlags.Fields) != MemberFlags.None)
				{
					FieldInfo[] fieldArr = this.EditedType.GetFields(BindingFlags.Instance | BindingFlags.Public);
					foreach (FieldInfo field in fieldArr)
					{
						PropertyEditor e = this.ParentGrid.PropertyEditorProvider.CreateEditor(field.FieldType, this, this.ParentGrid);
						e.Getter = this.CreateFieldValueGetter(field);
						e.Setter = this.CreateFieldValueSetter(field);
						e.PropertyName = field.Name;
						this.AddPropertyEditor(e);
					}
				}
				this.EndUpdate();

				this.Header.ExpandEnabled = this.PropertyEditors.Any();
				this.PerformGetValue();
			}
		}

		public override void PerformGetValue()
		{
			base.PerformGetValue();
			object[] values = this.Getter().ToArray();

			this.Header.ResetEnabled = !this.ReadOnly;

			if (values == null)
			{
				this.Header.ValueText = null;
				return;
			}

			string valString = null;
			if (!values.Any() || values.All(o => o == null))
			{
				this.ClearContent();

				this.Header.ExpandEnabled = false;
				this.Expanded = false;
				this.Header.ResetIsInit = true;
					
				valString = "null";
			}
			else
			{
				this.Header.ExpandEnabled = !this.ContentInitialized || this.PropertyEditors.Any();
				if (!this.Header.ExpandEnabled) this.Expanded = false;
				this.Header.ResetIsInit = false;

				valString = values.Count() == 1 ? 
					values.First().ToString() :
					string.Format(CustomPropertyGrid.Properties.Resources.PropertyGrid_N_Objects, values.Count());
			}

			foreach (PropertyEditor e in this.PropertyEditors)
				e.PerformGetValue();

			this.Header.ValueText = valString;
		}
		public override void PerformSetValue()
		{
			base.PerformSetValue();
			if (!this.PropertyEditors.Any()) return;

			foreach (PropertyEditor e in this.PropertyEditors)
				e.PerformSetValue();
		}

		protected Func<IEnumerable<object>> CreatePropertyValueGetter(PropertyInfo property)
		{
			return () => this.Getter().Select(o => o != null ? property.GetValue(o, null) : null);
		}
		protected Func<IEnumerable<object>> CreateFieldValueGetter(FieldInfo field)
		{
			return () => this.Getter().Select(o => o != null ? field.GetValue(o) : null);
		}
		protected Action<IEnumerable<object>> CreatePropertyValueSetter(PropertyInfo property)
		{
			return delegate(IEnumerable<object> values)
			{
				IEnumerator<object> valuesEnum = values.GetEnumerator();
				object[] targetArray = this.Getter().ToArray();

				object curValue = null;
				if (valuesEnum.MoveNext()) curValue = valuesEnum.Current;
				foreach (object target in targetArray)
				{
					if (target != null) property.SetValue(target, curValue, null);
					if (valuesEnum.MoveNext()) curValue = valuesEnum.Current;
				}

				// Fixup struct values by assigning the modified struct copy to its original member
				if (this.EditedType.IsValueType) this.Setter(targetArray);
			};
		}
		protected Action<IEnumerable<object>> CreateFieldValueSetter(FieldInfo field)
		{
			return delegate(IEnumerable<object> values)
			{
				IEnumerator<object> valuesEnum = values.GetEnumerator();
				object[] targetArray = this.Getter().ToArray();

				object curValue = null;
				if (valuesEnum.MoveNext()) curValue = valuesEnum.Current;
				foreach (object target in targetArray)
				{
					if (target != null) field.SetValue(target, curValue);
					if (valuesEnum.MoveNext()) curValue = valuesEnum.Current;
				}

				// Fixup struct values by assigning the modified struct copy to its original member
				if (this.EditedType.IsValueType) this.Setter(targetArray);
			};
		}

		private void Header_ResetClicked(object sender, EventArgs e)
		{
			if (this.EditedType.IsValueType)
			{
				this.SetterSingle(ReflectionHelper.CreateInstanceOf(this.EditedType));
			}
			else
			{
				if (this.Header.ResetIsInit)
				{
					this.SetterSingle(ReflectionHelper.CreateInstanceOf(this.EditedType));
					this.Expanded = true;
				}
				else
				{
					this.SetterSingle(null);
				}
			}

			this.PerformGetValue();
		}
	}
}
